# Write-Host "Installing functions"
[string]$KeyValuePairSeparator = "|";
[string]$KeyValueSeparator = "=";

function Get-XmlNamespaceManager([xml]$XmlDocument, [string]$NamespaceURI = "")
{
    # If a Namespace URI was not given, use the Xml document's default namespace.
    if ([string]::IsNullOrEmpty($NamespaceURI)) { $NamespaceURI = $XmlDocument.DocumentElement.NamespaceURI }    
    
    # In order for SelectSingleNode() to actually work, we need to use the fully qualified node path along with an Xml Namespace Manager, so set them up.
    [System.Xml.XmlNamespaceManager]$xmlNsManager = New-Object System.Xml.XmlNamespaceManager($XmlDocument.NameTable)
    $xmlNsManager.AddNamespace("ns", $NamespaceURI)
    return ,$xmlNsManager        # Need to put the comma before the variable name so that PowerShell doesn't convert it into an Object[].
}

function Get-FullyQualifiedXmlNodePath([string]$NodePath, [string]$NodeSeparatorCharacter = '.')
{
    return "/ns:$($NodePath.Replace($($NodeSeparatorCharacter), '/ns:'))"
}

function Get-XmlNode([xml]$XmlDocument, [string]$NodePath, [string]$NamespaceURI = "", [string]$NodeSeparatorCharacter = '.', [bool]$IsAppsettingsXml=$true)
{
    $xmlNsManager = Get-XmlNamespaceManager -XmlDocument $XmlDocument -NamespaceURI $NamespaceURI
    [string]$fullyQualifiedNodePath = Get-FullyQualifiedXmlNodePath -NodePath $NodePath -NodeSeparatorCharacter $NodeSeparatorCharacter
    
    # Try and get the node, then return it. Returns $null if the node was not found.
    $node = $XmlDocument.SelectSingleNode($fullyQualifiedNodePath, $xmlNsManager)
    return $node
}

function Get-XmlNodes([xml]$XmlDocument, [string]$NodePath, [string]$NamespaceURI = "", [string]$NodeSeparatorCharacter = '.', [bool]$IsAppsettingsXml=$true)
{
    $xmlNsManager = Get-XmlNamespaceManager -XmlDocument $XmlDocument -NamespaceURI $NamespaceURI
    [string]$fullyQualifiedNodePath = Get-FullyQualifiedXmlNodePath -NodePath $NodePath -NodeSeparatorCharacter $NodeSeparatorCharacter

    # Try and get the nodes, then return them. Returns $null if no nodes were found.
    # $nodes = $XmlDocument.SelectNodes($NodePath, $xmlNsManager)
    $nodes = $XmlDocument.SelectNodes($fullyQualifiedNodePath, $xmlNsManager)
    return $nodes
}

function Get-XmlElementsTextValue([xml]$XmlDocument, [string]$ElementPath, [string]$NamespaceURI = "", [string]$NodeSeparatorCharacter = '.')
{
    # Try and get the node.    
    $node = Get-XmlNode -XmlDocument $XmlDocument -NodePath $ElementPath -NamespaceURI $NamespaceURI -NodeSeparatorCharacter $NodeSeparatorCharacter
    
    # If the node already exists, return its value, otherwise return null.
    if ($node) { return $node.InnerText } else { return $null }
}

function Set-XmlElementsTextValue([xml]$XmlDocument, [string]$ElementPath, [string]$TextValue, [string]$NamespaceURI = "", [string]$NodeSeparatorCharacter = '.')
{
    # Try and get the node.    
    $node = Get-XmlNode -XmlDocument $XmlDocument -NodePath $ElementPath -NamespaceURI $NamespaceURI -NodeSeparatorCharacter $NodeSeparatorCharacter

    # If the node already exists, update its value.
    if ($node)
    { 
        $node.InnerText = $TextValue
    }
    # Else the node doesn't exist yet, so create it with the given value.
    else
    {
        # Create the new element with the given value.
        $elementName = $ElementPath.Substring($ElementPath.LastIndexOf($NodeSeparatorCharacter) + 1)
        $element = $XmlDocument.CreateElement($elementName, $XmlDocument.DocumentElement.NamespaceURI)        
        $textNode = $XmlDocument.CreateTextNode($TextValue)
        $element.AppendChild($textNode) > $null
        
        # Try and get the parent node.
        $parentNodePath = $ElementPath.Substring(0, $ElementPath.LastIndexOf($NodeSeparatorCharacter))
        $parentNode = Get-XmlNode -XmlDocument $XmlDocument -NodePath $parentNodePath -NamespaceURI $NamespaceURI -NodeSeparatorCharacter $NodeSeparatorCharacter
        
        if ($parentNode)
        {
            $parentNode.AppendChild($element) > $null
        }
        else
        {
            throw "$parentNodePath does not exist in the xml."
        }
    }
}

function Get-XmlElementsAttributeValue([ xml ]$XmlDocument, [string]$ElementPath, [string]$AttributeName, [string]$NamespaceURI = "", [string]$NodeSeparatorCharacter = '.')
{
    # Try and get the node. 
    $node = Get-XmlNode -XmlDocument $XmlDocument -NodePath $ElementPath -NamespaceURI $NamespaceURI -NodeSeparatorCharacter $NodeSeparatorCharacter
     
    # If the node already exists, return its value, otherwise return null.
    if ($node -and $node.$AttributeName) { return $node.$AttributeName } else { return $null }
}

function Set-XmlElementsAttributeValue([ xml ]$XmlDocument, [string]$ElementPath, [string]$AttributeName, [string]$AttributeValue, [string]$NamespaceURI = "", [string]$NodeSeparatorCharacter = '.')
{
    # Try and get the node. 
    $node = Get-XmlNode -XmlDocument $XmlDocument -NodePath $ElementPath -NamespaceURI $NamespaceURI -NodeSeparatorCharacter $NodeSeparatorCharacter
     
    # If the node already exists, create/update its attribute's value.
    if ($node)
    { 
        $attribute = $XmlDocument.CreateNode([System.Xml.XmlNodeType]::Attribute, $AttributeName, $NamespaceURI)
        $attribute.Value = $AttributeValue
        $node.Attributes.SetNamedItem($attribute) > $null
    }
    # Else the node doesn't exist yet, so create it with the given attribute value.
    else
    {
        # Create the new element with the given value.
        $elementName = $ElementPath.SubString($ElementPath.LastIndexOf($NodeSeparatorCharacter) + 1)
        $element = $XmlDocument.CreateElement($elementName, $XmlDocument.DocumentElement.NamespaceURI)
        $element.SetAttribute($AttributeName, $NamespaceURI, $AttributeValue) > $null
         
        # Try and get the parent node.
        $parentNodePath = $ElementPath.SubString(0, $ElementPath.LastIndexOf($NodeSeparatorCharacter))
        $parentNode = Get-XmlNode -XmlDocument $XmlDocument -NodePath $parentNodePath -NamespaceURI $NamespaceURI -NodeSeparatorCharacter $NodeSeparatorCharacter
        
        if ($parentNode)
        {
            $parentNode.AppendChild($element) > $null
        }
        else
        {
            throw "$parentNodePath does not exist in the xml."
        }
    }
}

function Set-XmlElementsAttributeValueOnNode([ System.Xml.XmlNode ]$XmlNode, [string]$AttributeName, [string]$AttributeValue, [string]$NamespaceURI = "", [string]$NodeSeparatorCharacter = '.')
{
    # If the node already exists, create/update its attribute's value.
    if ($XmlNode)
    { 
        $attribute = $XmlDocument.CreateNode([System.Xml.XmlNodeType]::Attribute, $AttributeName, $NamespaceURI)
        $attribute.Value = $AttributeValue
        $XmlNode.Attributes.SetNamedItem($attribute) > $null
    }
}

function Get-XPath([System.Xml.XmlNode]$node, [string]$NodeSeparatorCharacter = '.') {
    if ( $node.GetType().Name -ne 'XmlDocument' ) {
        $parentNode = Get-XPath -node $node.ParentNode;
        if ( $null -eq $parentNode)
        {
            return $node.Name;
        }

        $nodeName = $node.Name;
        $elementName = $node | ForEach-Object { $_.GetType().GetProperty("Name").GetValue($_, $null); }
        if ($nodeName -cmatch '^\d+$')
        {
            $nodeName = $nodeName -creplace '(^\d+$)', '[$1]'
            $nodeName = $elementName + $nodeName;
        }
        return "{0}$NodeSeparatorCharacter{1}" -f ($parentNode), $nodeName
    }
}

Function Get-ObjectProperty
{
  param
  (
    $Name = '*',
    $Value = '*',
    $Type = '*',
    [Switch]$IsNumeric,
 
    [Parameter(Mandatory=$true,ValueFromPipeline=$true)]
    [Object[]]$InputObject,
 
    $Depth = 4,
    $Prefix = '$obj'
  )
 
  Begin
  {
    $x = 0
    Function Get-Property
    {
      param
      (
        $Node,
        [String[]]$Prefix
      )
 
      $Value = @{Name='Value'; Expression={$_.'#text' }}
      Select-Xml -Xml $Node -XPath 'Property' | ForEach-Object {$i=0} {
        $rv = $_.Node | Select-Object -Property Name, $Value, Path, Type
        $isCollection = $rv.Name -eq 'Property'
 
        if ($isCollection)
        {
          $CollectionItem = "[$i]"
          $i++
          $rv.Path = (($Prefix) -join '.') + $CollectionItem
        }
        else
        {
          $rv.Path = ($Prefix + $rv.Name) -join '.'
        }
 
        $rv
 
        if (Select-Xml -Xml $_.Node -XPath 'Property')
        {
          if ($isCollection)
          {
            $PrefixNew = $Prefix.Clone()
            $PrefixNew[-1] += $CollectionItem
            Get-Property -Node $_.Node -Prefix ($PrefixNew )
          }
          else
          {
            Get-Property -Node $_.Node -Prefix ($Prefix + $_.Node.Name )
          }
        }
      }
    }
  }
 
  Process
  {
    $x++
    $InputObject |
    ConvertTo-Xml -Depth $Depth |
    ForEach-Object { $_.Objects } |
    ForEach-Object { Get-Property $_.Object -Prefix $Prefix$x  } |
    Where-Object { $_.Name -like "$Name" } |
    Where-Object { $_.Value -like $Value } |
    Where-Object { $_.Type -like $Type } |
    Where-Object { $IsNumeric.IsPresent -eq $false -or $_.Value -as [Double] }
  }
}

function Get-ConfigControls([xml]$XmlDocument, [string]$NodePath, [bool]$OnlyWithOrderAttribute=$true)
{
    $nodes = Get-XmlNodes -XmlDocument $XmlDocument -NodePath $NodePath | Sort-Object -Property name

    # Child Nodes
    ForEach ($node in $nodes.ChildNodes)
    {
        $currentNodePath = Get-XPath($node);
        $elementName = $node | ForEach-Object { $_.GetType().GetProperty("Name").GetValue($_, $null); }
        
        [bool] $isArray = $false;
        try {
            $nameAttr = Get-XmlElementsAttributeValue -XmlDocument $AppSettingsXml -ElementPath $currentNodePath -AttributeName "name"
        }
        catch {
            $isArray = $true;
            # Fix $currentNodePath
            $parentNodePath = $currentNodePath.SubString(0, $currentNodePath.LastIndexOf('.'))
            $currentNodePath = "$parentNodePath.$elementName"
        }

        if ($nameAttr -ne $elementName)
        {
            $isArray = $true;
        }

        $orderAttr = Get-XmlElementsAttributeValue -XmlDocument $AppSettingsXml -ElementPath $currentNodePath -AttributeName "order"
        $displayNameAttr = Get-XmlElementsAttributeValue -XmlDocument $AppSettingsXml -ElementPath $currentNodePath -AttributeName "display_name"
        $descriptionAttr = Get-XmlElementsAttributeValue -XmlDocument $AppSettingsXml -ElementPath $currentNodePath -AttributeName "descr"
        $controlTypeAttr = Get-XmlElementsAttributeValue -XmlDocument $AppSettingsXml -ElementPath $currentNodePath -AttributeName "control_type"
        $allowedValuesAttr = Get-XmlElementsAttributeValue -XmlDocument $AppSettingsXml -ElementPath $currentNodePath -AttributeName "allowed_values"
        $installerAttr = Get-XmlElementsAttributeValue -XmlDocument $AppSettingsXml -ElementPath $currentNodePath -AttributeName "installer"
        $nodeText = Get-XmlElementsTextValue -XmlDocument $AppSettingsXml -ElementPath $currentNodePath
        
        # [bool] $isArray = $null -ne $rowAttr
        # Write-Output "NAME :::: $nameAttr"
        # Write-Output "ELEMENT NAME :::: $elementName"
        
        [bool] $isGroup = $controlTypeAttr -eq "group";
        [bool] $isControl = $null -ne $controlTypeAttr -and $isGroup -ne $true;
        
        if ($null -eq $orderAttr) { 
            if ($OnlyWithOrderAttribute -eq $true -and $isArray -eq $false)
            { 
                # Hide all elements without order attrib
                # Unless it's an array element
                continue; 
            }
        }

        Write-Output ""
        Write-Output "-------- $elementName :: $currentNodePath --------"
        Write-Output ""

        if ($isGroup)
        {
            Write-Output "GROUP:"
            Write-Output "Display Name: $displayNameAttr";
            Write-Output "Description: $descriptionAttr";
            Write-Output ""
        }
        elseif ($isControl){
            Write-Output "CONTROL:"
            Write-Output "Display Name: $displayNameAttr";
            Write-Output "Description: $descriptionAttr";
            Write-Output "ControlType: $controlTypeAttr";
            Write-Output "Installer: $installerAttr";
            Write-Output "Allowed Values: $allowedValuesAttr";
            Write-Output "Text: $nodeText";
            Write-Output ""
        }
        elseif ($isArray)
        {
            Write-Output "ARRAY: $rowAttr"
            Write-Output "Path: $currentNodePath";
            
            $pathWithoutIndex = $currentNodePath -creplace '\[\d+\]+$', '' # Only affect last part of path (ends with)
            Write-Output "Path WO Index: $pathWithoutIndex";
            $arrayNodes = Get-XmlNodes -XmlDocument $AppSettingsXml -NodePath $pathWithoutIndex | Sort-Object -Property name
            $arrCount = $arrayNodes.Count;
            Write-Output "Total Count: $arrCount"
            # $i = 0;
            # Iterate through array elements
            ForEach ($arrayNode in $arrayNodes)
            {
                $rowIndex = $arrayNode.name;
                Write-Output "ROW _________ INDEX :: $rowIndex"
                if ($null -eq $rowIndex) { continue; }

                if ("0" -eq $rowIndex)
                {
                    Write-Error "Configuration arrays name attribute must start at 1, not 0!"
                }

                $indexedNodePath = "$pathWithoutIndex[$rowIndex]";
                # $indexedNodePath = "$pathWithoutIndex[$i]";

                $nodeText = Get-XmlElementsTextValue -XmlDocument $AppSettingsXml -ElementPath $indexedNodePath
                Write-Output "Text: $nodeText";
                
                Get-ConfigControls -XmlDocument $AppSettingsXml -NodePath $indexedNodePath -OnlyWithOrderAttribute $OnlyWithOrderAttribute;
            }

            return;
        }

        # $childRowAttr = Get-XmlElementsAttributeValue -XmlDocument $AppSettingsXml -ElementPath $currentNodePath -AttributeName "row"
        if ($isArray -eq $false -and $node.ChildNodes.NodeType -contains 'Element')
        {
            Write-Output "Has ELEMENTS ($currentNodePath)"
            Get-ConfigControls -XmlDocument $AppSettingsXml -NodePath $currentNodePath -OnlyWithOrderAttribute $OnlyWithOrderAttribute;
        }
    }
}

function Get-ConfigurableElementPaths([xml]$XmlDocument, [string]$NodePath, [bool]$OnlyWithOrderAttribute=$true)
{
    $result = New-Object System.Collections.Generic.List[System.String];

    $nodes = Get-XmlNodes -XmlDocument $XmlDocument -NodePath $NodePath | Sort-Object -Property name

    # Child Nodes
    ForEach ($node in $nodes.ChildNodes)
    {
        $currentNodePath = Get-XPath($node);

        $elementName = $node | ForEach-Object { $_.GetType().GetProperty("Name").GetValue($_, $null); }
        
        [bool] $isArray = $false;
        try {
            $nameAttr = Get-XmlElementsAttributeValue -XmlDocument $AppSettingsXml -ElementPath $currentNodePath -AttributeName "name"
        }
        catch {
            $isArray = $true;
            # Fix $currentNodePath
            $parentNodePath = $currentNodePath.SubString(0, $currentNodePath.LastIndexOf('.'))
            $currentNodePath = "$parentNodePath.$elementName"
        }

        if ($nameAttr -ne $elementName)
        {
            $isArray = $true;
        }

        $orderAttr = Get-XmlElementsAttributeValue -XmlDocument $AppSettingsXml -ElementPath $currentNodePath -AttributeName "order"
        $controlTypeAttr = Get-XmlElementsAttributeValue -XmlDocument $AppSettingsXml -ElementPath $currentNodePath -AttributeName "control_type"
        
        [bool] $isGroup = $controlTypeAttr -eq "group";
        [bool] $isControl = $null -ne $controlTypeAttr -and $isGroup -ne $true;

        if ($null -eq $orderAttr) { 
            if ($OnlyWithOrderAttribute -eq $true -and $isArray -eq $false)
            { 
                # Hide all elements without order attrib
                # Unless it's an array element
                continue; 
            }
        }

        if ($isGroup)
        {
            # Write-Output "GROUP:"
        }
        elseif ($isControl){
            # Write-Output "CONTROL:"
            # Write-Output $currentNodePath;
            $result.Add($currentNodePath);
        }
        elseif ($isControl -eq $false -and $isArray -eq $false -and $OnlyWithOrderAttribute -eq $false)
        {
            $result.Add($currentNodePath);
        }
        elseif ($isArray)
        {
            # Write-Output $currentNodePath;
            
            $pathWithoutIndex = $currentNodePath -creplace '\[\d+\]+$', '' # Only affect last part of path (ends with)

            $arrayNodes = Get-XmlNodes -XmlDocument $AppSettingsXml -NodePath $pathWithoutIndex | Sort-Object -Property name
            # $arrCount = $arrayNodes.Count;
            # Write-Output "Total Count: $arrCount"
            
            # Iterate through array elements
            ForEach ($arrayNode in $arrayNodes)
            {
                $rowIndex = $arrayNode.name;
                if ($null -eq $rowIndex) { continue; }

                if ("0" -eq $rowIndex)
                {
                    Write-Error "Configuration arrays name attribute must start at 1, not 0!"
                }

                $indexedNodePath = "$pathWithoutIndex[$rowIndex]";
                # $indexedNodePath = "$pathWithoutIndex[$i]";
                
                Get-ConfigurableElementPaths -XmlDocument $AppSettingsXml -NodePath $indexedNodePath -OnlyWithOrderAttribute $OnlyWithOrderAttribute;
            }

            return;
        }

        if ($isArray -eq $false -and $node.ChildNodes.NodeType -contains 'Element')
        {
            # Write-Output "Has ELEMENTS ($currentNodePath)"
            Get-ConfigurableElementPaths -XmlDocument $AppSettingsXml -NodePath $currentNodePath -OnlyWithOrderAttribute $OnlyWithOrderAttribute;
        }
    }

    return $result;
}

function Set-ConfigurationValue
{
    [CmdletBinding()]
    Param (
        [Parameter(Mandatory = $true)]
        [xml]
        $XmlDocument,
        
        [Parameter(Mandatory = $true)]
        [string]
        $ConfigurationKey,
        
        [Parameter(Mandatory = $true)]
        [AllowEmptyString()]
        [string]
        $ConfigurationValue
    )

    $configRootPath = Get-AppsettingsXPath([string]::empty);
    $configKeys = Get-ConfigurableElementPaths -XmlDocument $XmlDocument -NodePath $configRootPath -OnlyWithOrderAttribute $false;

    $isArray = $false;
    $arrayBracketRegex = [regex] '(\[\d+\])'
    if ($arrayBracketRegex.IsMatch($ConfigurationKey)) {
        $isArray = $true;
    }

    $keyExistsInAppSettings = $true;
    if (-not $configKeys.Contains($ConfigurationKey))
    {
        $keyExistsInAppSettings = $false;
    }
    
    if ($keyExistsInAppSettings -eq $true)
    {
        if ($isArray)
        {
            #Update array value. Do we really need to check if isArray?
        }

        $pscmdlet.WriteDebug("Updating: $ConfigurationKey to new value: $ConfigurationValue");
        Set-XmlElementsTextValue -XmlDocument $XmlDocument -ElementPath $ConfigurationKey -TextValue $ConfigurationValue
    }
    
    if ($keyExistsInAppSettings -eq $false)
    {
        # configuration.ServiceInformation.HostingAddresses.HostingAddress[1].Url
        # Future nesting
        # configuration.ServiceInformation.HostingAddresses.HostingAddress[3].HostingAddresses2.HostingAddress2[2].Url

        if (-not $isArray) {
            $pscmdlet.WriteDebug("The configuration XPath $ConfigurationKey does not exist in the configuration file.");
            return $false;
        }

        # Get first configuration key (index 1 on all levels)
        $firstIndexConfigurationKey = $arrayBracketRegex.Replace($ConfigurationKey, '[1]')
        # Check if the levelone[1].leveltwo[1].value path exists in the config
        $baseNodes = Get-XmlNodes -XmlDocument $AppSettingsXml -NodePath $firstIndexConfigurationKey;
        if ($null -eq $baseNodes)
        {
            $pscmdlet.WriteDebug("The configuration XPath $firstIndexConfigurationKey does not exist in the configuration file. Having a default value with name=1 for all array levels is mandatory.");
            return $false;
        }

        # Iterate over matches
        $matchdetails = $arrayBracketRegex.Match($ConfigurationKey)
        # $matchCount = $matchdetails.Groups.Count;
        while ($matchdetails.Success)
        {
            $arrivedAtLastNode = $false;
            for ($i = 1; $i -lt $matchdetails.Groups.Count; $i++)
            {
                if ($i -eq $matchdetails.Groups.Count)
                {
                    $arrivedAtLastNode = $true;
                    $pscmdlet.WriteDebug("Arrived at last: $arrivedAtLastNode ($i)");
                }

                $groupdetails = $matchdetails.Groups[$i]
                if ($groupdetails.Success) {
                    $matchedText = $groupdetails.Value;
                    # Get $newNameAttrValue [8] = 8
                    $numberRegex = [regex] '\d+'
                    $newNameAttrValue = $numberRegex.Match($matchedText).Value;

                    $matchStart = $groupdetails.Index;
                    $matchLength = $groupdetails.Length;
                    
                    $arrayIndexPath = $ConfigurationKey.Substring(0, $matchStart + $matchLength);
                    $arrayBasePath = $ConfigurationKey.Substring(0, $matchStart);
                    
                    # Check if basepath (array) exists in config - error if not.
                    $baseNodes = Get-XmlNodes -XmlDocument $AppSettingsXml -NodePath $arrayBasePath;
                    if ($null -eq $baseNodes)
                    {
                        $pscmdlet.WriteDebug("The path $arrayBasePath does not exist in the appsettings file. Skipping setting.");
                        return $false;
                    }

                    # Check if array member exists at this level - if not, create it from the one with index 1
                    $baseNodes = Get-XmlNodes -XmlDocument $AppSettingsXml -NodePath $arrayIndexPath;
                    if ($null -eq $baseNodes)
                    {
                        # CREATE Case
                        $parentNodePath = $arrayBasePath.Substring(0, $arrayBasePath.LastIndexOf("."));
                        # Clone node with index 1
                        # Start by getting the same path from $firstIndexConfigurationKey
                        $fMatchDetails = $arrayBracketRegex.Match($firstIndexConfigurationKey);
                        $fGroupdetails = $fMatchdetails.Groups[$i];
                        
                        $fMatchedText = $fGroupdetails.Value;
                        $fMatchStart = $fGroupdetails.Index;
                        $fMatchLength = $fGroupdetails.Length;
                        
                        $fArrayIndexPath = $firstIndexConfigurationKey.Substring(0, $fMatchStart + $fMatchLength);
                        $fArrayBasePath = $firstIndexConfigurationKey.Substring(0, $fMatchStart);
                        $pscmdlet.WriteDebug("CREATE: $fMatchedText, Base: $fArrayBasePath, Parent: $parentNodePath, Index: $i");
                        
                        $NodeToClone = Get-XmlNode -XmlDocument $AppSettingsXml -NodePath $fArrayIndexPath;
                        $ClonedNode = $NodeToClone.Clone();

                        # Update node name attribute to $newNameAttrValue
                        Set-XmlElementsAttributeValueOnNode -XmlNode $ClonedNode -AttributeName "name" -AttributeValue $newNameAttrValue
                        
                        # Add cloned node to ParentNode
                        $pscmdlet.WriteDebug("PNP: $parentNodePath ($fArrayIndexPath)");
                        $ParentNode = Get-XmlNode -XmlDocument $AppSettingsXml -NodePath $parentNodePath;
                        $ParentNode.AppendChild($ClonedNode);

                        # Next: recurse
                    }
                } 
            }
            $matchdetails = $matchdetails.NextMatch();
        }

        # Recurse
        return Set-ConfigurationValue -XmlDocument $XmlDocument -ConfigurationKey $ConfigurationKey -ConfigurationValue $ConfigurationValue
    }
    
    return $true;
}

function Get-AppsettingsXPath([string]$NodeOrElementPath)
{
    if ($NodeOrElementPath -eq [string]::empty)
    {
        $NodeOrElementPath = "configuration"
    }
    elseif (-not ($NodeOrElementPath.StartsWith("configuration.")))
    {
        $NodeOrElementPath = "configuration.$NodeOrElementPath"
    }

    return $NodeOrElementPath
}

function Set-ConfigStringKeyPrefixes
{
    [CmdletBinding()]
    Param (
        [Parameter(Mandatory = $true)]
        [AllowEmptyString()]
        [string]
        $ConfigString
    )

    If($ConfigString -eq [string]::empty)
    {
        return [string]::empty;
    }

    $NewConfigString = [string]::empty;

    $configKvps = $ConfigString.Split($KeyValuePairSeparator)
    foreach ($newKvp in $configKvps) 
    {
        $newKey   = Get-AppsettingsXPath($newKvp.SubString(0, $newKvp.IndexOf($KeyValueSeparator)));
        $newValue = $newKvp.SubString($newKvp.IndexOf($KeyValueSeparator) + 1);
        $NewConfigString = Add-KvpToConfigString -ConfigString $NewConfigString -ConfigurationKey $newKey -ConfigurationValue $newValue
    }

    return $NewConfigString;
}

function Set-UpdateAndSaveXmlFile
{
    [CmdletBinding()]
    Param (
        [Parameter(Mandatory = $true)]
        [xml]
        $XmlDocument,
        
        [Parameter(Mandatory = $true)]
        [AllowEmptyString()]
        [string]
        $ConfigurationString,

        [Parameter(Mandatory = $true)]
        [string]
        $XmlFilePath
    )

    $newConfigKvps = $ConfigurationString.Split($KeyValuePairSeparator)
    $pscmdlet.WriteDebug("Configuration: $ConfigurationString");
    foreach ($newKvp in $newConfigKvps) {
        $newKey   = $newKvp.SubString(0, $newKvp.IndexOf($KeyValueSeparator))
        $newValue = $newKvp.SubString($newKvp.IndexOf($KeyValueSeparator) + 1)
        Set-ConfigurationValue -XmlDocument $XmlDocument -ConfigurationKey $newKey -ConfigurationValue $newValue > $null; # Do not output result (bool)
    }

    $pscmdlet.WriteDebug("Saving xml to: $XmlFilePath");
    $AppSettingsXml.Save($XmlFilePath);
}

function Test-ConfigStringContainsKey
{
    [CmdletBinding()]
    Param (
        [Parameter(Mandatory = $true)]
        [AllowEmptyString()]
        [string]
        $ConfigString,
        
        [Parameter(Mandatory = $true)]
        [string]
        $ConfigurationKey
    )

    If($ConfigString -eq [string]::empty)
    {
        return $false
    }

    $configKvps = $ConfigString.Split($KeyValuePairSeparator)
    foreach ($newKvp in $configKvps) {
        $newKey = $newKvp.SubString(0, $newKvp.IndexOf($KeyValueSeparator))
        if ($newKey -eq $ConfigurationKey)
        {
            return $true;
        }
    }

    return $false;
}
function Add-KvpToConfigString
{
    [CmdletBinding()]
    Param (
        [Parameter(Mandatory = $true)]
        [AllowEmptyString()]
        [string]
        $ConfigString,
        
        [Parameter(Mandatory = $true)]
        [string]
        $ConfigurationKey,
        
        [Parameter(Mandatory = $true)]
        [AllowEmptyString()]
        [string]
        $ConfigurationValue
    )

    if ($ConfigString -eq [string]::empty)
    {
        $ConfigString = "$ConfigurationKey=$ConfigurationValue"
    }

    $alreadySupplied = Test-ConfigStringContainsKey -ConfigString $ConnectorConfig -ConfigurationKey $ConfigurationKey
    if ($alreadySupplied -eq $false)
    {
        $ConfigString = "$ConfigString|$ConfigurationKey=$ConfigurationValue"
    }

    return $ConfigString;
}

function Test-IsGuid
{
    [OutputType([bool])]
    param
    (
        [Parameter(Mandatory = $true)]
        [string]$ObjectGuid
    )

    # Define verification regex
    [regex]$guidRegex = '(?im)^[{(]?[0-9A-F]{8}[-]?(?:[0-9A-F]{4}[-]?){3}[0-9A-F]{12}[)}]?$'

    # Check guid against regex
    return $ObjectGuid -match $guidRegex
}

function Test-IsGuidEmpty
{
    [OutputType([bool])]
    param
    (
        [Parameter(Mandatory = $true)]
        [string]$ObjectGuid
    )

    If(Test-IsGuid($ObjectGuid))
    {
        return $ObjectGuid -eq [system.guid]::empty
    }
    return $false; 
}

# Write-Host "Functions Installed"