﻿#Requires -RunAsAdministrator

<#
.SYNOPSIS
   Install CaaS and apply configuration 
.EXAMPLE
    .\winsvc-install.ps1 
    -ConnectorVersion "0.0.12"
    -ConnectorId "00000000-0000-0000-0000-000000000000" 
    -PortNumber 58080
    -ConnectorConfig "DefaultServers.DefaultServer[1].Url=http://rndqa1.novacura.local/master/NovaCura.Flow.Server/|DefaultServers.DefaultServer[1].User=flowUser1|DefaultServers.DefaultServer[1].Password=pwd1"
   -Verbose
#>


[CmdletBinding(SupportsShouldProcess=$true, PositionalBinding=$false)]
Param(
    [int]$DefaultPortNumber = 58080,
    [string]$ConnectorName = [string]::empty,
    [string]$SourceFolderName = [string]::empty,
    [string]$ConnectorDisplayName = [string]::empty,
    [int]$StartService = 1,

#------------------------------------------------------------------------------------------------------------------
    [Parameter(Mandatory=$false, HelpMessage="Connector Service version")]
    [string]$ConnectorVersion = [string]::empty,
    
    [Parameter(Mandatory=$false, HelpMessage="Service Name")]
    [string]$ServiceName = [string]::empty,

    [Parameter(Mandatory=$false, HelpMessage="Service Description")]
    [string]$ServiceDescription = [string]::empty,
    
    [Parameter(Mandatory=$false, HelpMessage="Destination base path for connector files.")]
    [string]$DestPath = "C:\Novacura Connectors",
    
    [Parameter(Mandatory=$false, HelpMessage="Port for the connector service. Default: 58080")]
    [int]$PortNumber = $DefaultPortNumber,
    
    [Parameter(Mandatory=$false, HelpMessage="Connector Service ID.")]
    [string]$ConnectorId = "",
    
    [Parameter(Mandatory=$false, HelpMessage="Connector configuration (key(xpath case sensitive)=value|key(xpath)=value)")]
    [string]$ConnectorConfig = [string]::empty
)

function Use-64Bit {
    [CmdletBinding()]
    [OutputType([Bool])]
    Param (
        [parameter(Mandatory = $True)]
        [String]
        $PsRootPath,
        [parameter(Mandatory = $True)]
        [String]
        $ScriptRootPath,
        [parameter(Mandatory = $True)]
        [String]
        $ScriptFilePath
    )
    Process {
        # Are you running in 32-bit mode?
        #   (\SysWOW64\ = 32-bit mode)
        if ($PsRootPath -like "*SysWOW64*") {
            Write-Warning "Restarting this script under 64-bit Windows PowerShell."

            # Restart this script under 64-bit Windows PowerShell.
            #   (\SysNative\ redirects to \System32\ for 64-bit mode)
            & (Join-Path ($PsRootPath -replace "SysWOW64", "SysNative") powershell.exe) `
                -NonInteractive `
                -NoProfile `
                -File (Join-Path $ScriptRootPath ([System.IO.Path]::GetFileName($ScriptFilePath))) `
                2>&1 | Tee-Object -Variable allOutput

            Write-Output ($allOutput | Format-List | Out-String)

            return $TRUE
        }
        return $FALSE
    }
}

# Force running script in 64-bit mode
$32bitRun = Use-64Bit -PsRootPath $PSHOME -ScriptRootPath $PSScriptRoot -ScriptFilePath $MyInvocation.MyCommand
if ($32BitRun -eq $TRUE) { exit $LastExitCode }

# region Include required files
$ScriptDirectory = Split-Path -Path $MyInvocation.MyCommand.Definition -Parent
Push-Location $ScriptDirectory
try {
    . ("$ScriptDirectory\_include.ps1")
}
catch {
    Write-Host "Error while loading supporting PowerShell Scripts" 
}
#endregion

# Get SourceFolderName
if ($SourceFolderName -eq [string]::empty)
{
    $folders = Get-ChildItem -Directory -Filter "Novacura.Connector*"| Select-Object Name
    $SourceFolderName = $folders[0].Name
    Write-Output "SourceFolderName: $SourceFolderName"
}

# Get ConnectorName (SourceFolderName minus "Novacura.Connector.")
if ($ConnectorName -eq [string]::empty)
{
    $ConnectorName = $SourceFolderName -ireplace [regex]::Escape("novacura.connector."), "";
}
else
{
    $ConnectorName = $ConnectorName -ireplace [regex]::Escape("novacura.connector."), "";
}
Write-Output "Connector Name: $ConnectorName"

# Load appsettings.xml
[ xml ]$AppSettingsXml = Get-Content -Path "$PSScriptRoot\$SourceFolderName\appsettings.xml"

# Get connector version
#$NewXmlConnectorVersionValue = $true;
$exeFilePath = "$PSScriptRoot\$SourceFolderName\$SourceFolderName.dll"
if ($ConnectorVersion -eq [string]::empty)
{
    # Use the version from the main assembly
    $FileMajorPart = [System.Diagnostics.FileVersionInfo]::GetVersionInfo($exeFilePath).FileMajorPart
    $FileMinorPart = [System.Diagnostics.FileVersionInfo]::GetVersionInfo($exeFilePath).FileMinorPart
    $FileBuildPart = [System.Diagnostics.FileVersionInfo]::GetVersionInfo($exeFilePath).FileBuildPart
    # $FilePrivatePart = [System.Diagnostics.FileVersionInfo]::GetVersionInfo($exeFilePath).FilePrivatePart
    $ConnectorVersion = "$FileMajorPart.$FileMinorPart.$FileBuildPart"

    # Exit if we failed to get the connector version
    if ($ConnectorVersion -eq [string]::empty)
    { 
        Write-Error "Could not get ConnectorVersion from $exeFilePath"
        return
    }
    
#    $NewXmlConnectorVersionValue = $false;
}

Write-Output "Connector Version: $ConnectorVersion"

# Set ConnectorDisplayName
$ConnectorDisplayName = "$ConnectorName v$ConnectorVersion @ $env:COMPUTERNAME"
Write-Output "Connector DisplayName: $ConnectorDisplayName"

# Set ServiceName
if ($ServiceName -eq [string]::empty)
{
    $ServiceName = "Novacura.Connector.$ConnectorName.v$ConnectorVersion"
    Write-Output "Connector ServiceName: $ServiceName"
}

# Generate a new guid if required
If($ConnectorId -eq [string]::empty)
{
    $ConnectorId = New-Guid
}
ElseIf((Test-IsGuidEmpty($ConnectorId)))
{
    $ConnectorId = New-Guid
}
ElseIf(!(Test-IsGuid($ConnectorId)))
{
    $ConnectorId = New-Guid
}

# Update appsettings.xml
#------------------------------------------------------------------------------------------------------------------
# List-Configuration-Controls -XmlDocument $AppSettingsXml -NodePath "configuration";

# Fix ConnectorConfig (prefix all settings with configuration.)
$ConnectorConfig = Set-ConfigStringKeyPrefixes($ConnectorConfig);

# Add ConnectorId to Configuration String:
$currentKey = Get-AppsettingsXPath("ServiceInformation.Id");
$ConnectorConfig = Add-KvpToConfigString -ConfigString $ConnectorConfig -ConfigurationKey $currentKey -ConfigurationValue $ConnectorId

# Add DisplayName:
$currentKey = Get-AppsettingsXPath("ServiceInformation.DisplayName");
$ConnectorConfig = Add-KvpToConfigString -ConfigString $ConnectorConfig -ConfigurationKey $currentKey -ConfigurationValue $ConnectorDisplayName

# Add HostingAddress (first array item)
$currentKey = Get-AppsettingsXPath("ServiceInformation.HostingAddresses.HostingAddress[1].Url");
$hostedAt = "http://"+$env:COMPUTERNAME.ToLower()+":" + $PortNumber
$ConnectorConfig = Add-KvpToConfigString -ConfigString $ConnectorConfig -ConfigurationKey $currentKey -ConfigurationValue $hostedAt

# Apply the ConnectorConfig to appsettings.xml and save the appsettings file before copying the files
Write-Output "Updating application settings..."
Set-UpdateAndSaveXmlFile -XmlDocument $AppSettingsXml -ConfigurationString $ConnectorConfig -XmlFilePath "$PSScriptRoot\$SourceFolderName\appsettings.xml"
#------------------------------------------------------------------------------------------------------------------

# Set ServiceDescription
$ServiceDescriptionXPath = Get-AppsettingsXPath("ServiceInformation.Description");
$ServiceDescription = Get-XmlElementsTextValue -XmlDocument $AppSettingsXml -ElementPath $ServiceDescriptionXPath
$ServiceDescription = "$ServiceDescription. Port: $PortNumber"
Write-Output $ServiceDescription


Write-Output "Connector Id: $ConnectorId"

$ServiceFolder = Join-Path $DestPath $ConnectorName
$ServiceFolder = Join-Path $ServiceFolder $ConnectorVersion

Write-Output "DestPath: $DestPath"
Write-Output "ServiceFolder: $ServiceFolder"

# Check if destination exists
If(!(Test-Path $DestPath))
{
    New-Item -ItemType Directory -Force -Path $DestPath
}
If(!(Test-Path $ServiceFolder))
{
   New-Item -ItemType Directory -Force -Path $ServiceFolder
}

# Set permission rights for CS back-end
# SID S-1-5-19 == "LOCAL SERVICE"
$localServiceSID = New-Object System.Security.Principal.SecurityIdentifier("S-1-5-19")
$localServiceName = ($localServiceSID.Translate( [System.Security.Principal.NTAccount])).Value
$AccessRule = New-Object System.Security.AccessControl.FileSystemAccessRule($localServiceName, "FullControl", "ContainerInherit, ObjectInherit", "None", "Allow")
try {
    $acl = Get-Acl -Path $ServiceFolder -ErrorAction Stop
    $acl.SetAccessRule($AccessRule)
    Set-Acl -Path $ServiceFolder -AclObject $acl -ErrorAction Stop
    Write-Host "Successfully set FullControl permissions on $ServiceFolder"
} catch {
    throw "$ServiceFolder : Failed to set permissions. Details : $_"
}

# Delete Temporary ServiceFolder folder
C:\Windows\System32\cmd.exe /c rd /s /q $ServiceFolder
# And copy the files
Write-Output "Copying files..."
Copy-Item -Path "$PSScriptRoot\$SourceFolderName" -Destination $ServiceFolder -Force -Recurse

# Install windows service
Write-Output "Installing the service..."
$params = @{
  Name = $ServiceName
  BinaryPathName = "$ServiceFolder\$SourceFolderName.exe --urls=$hostedAt"
  DisplayName = $ServiceName
  StartupType = "Automatic"
  Description = $ServiceDescription
}
New-Service @params

if ($StartService -eq 1)
{
    Write-Output "Starting the service..."
    Start-Service -Name $ServiceName   
}

Write-Output "Done!"
Write-Output "Connector Web Administration URL: $hostedAt"

Pop-Location

#Start $hostedAt